/*
 *   Copyright (C) 2008 Schutz Sacha <istdasklar@gmail.com>
 *   AND lex newton http://qgoogletranslate.googlecode.com
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2,
 *   or (at your option) any later version, as published by the Free
 *   Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.*/


#include "ktranslatoid.h"

#include <QClipboard>
#include <QLabel>
#include <QMenu>
#include <QPainter>
#include <QTextCodec>
#include <QToolButton>
#include <QTreeView>

#include <KConfigDialog>
#include <KDebug>
#include <KMessageBox>
#include <KPushButton>
#include <KStandardDirs>
#include <KTextEdit>
#include <KIconLoader>
#include <KIO/Job>

#include <Solid/Device>
#include <Solid/Networking>

#include <Plasma/Containment>
#include <Plasma/Corona>
#include <Plasma/Theme>

KTranslatoid ::KTranslatoid(QObject *parent, const QVariantList &args)
    :  Plasma::PopupApplet(parent, args),
    m_job(0),
    m_voiceForm(0)
{
    m_sourceText = new Plasma::TextEdit(this);
    m_destText = new Plasma::TextEdit(this);
    m_buttonTranslate = new Plasma::PushButton(this);
    m_buttonChange = new Plasma::ToolButton(this);
    m_sourceLanguageLabel = new Plasma::Label(this);
    m_destLanguageLabel = new Plasma::Label(this);
    m_sourceLanguageButton = new Plasma::ToolButton(this);
    m_destLanguageButton = new Plasma::ToolButton(this);
    m_popupIcon = KIcon("translator");
    m_autoPaste = true;
    m_autoTranslate = true;
    m_languageModel = new LanguageModel(this);

    m_sourceText->nativeWidget()->installEventFilter(this);
    connect(m_buttonTranslate, SIGNAL(clicked()), this, SLOT(translate()));
    connect(m_buttonChange, SIGNAL(clicked()), this, SLOT(swapLanguages()));
    connect(m_sourceLanguageButton, SIGNAL(clicked()), this, SLOT(changeSourceLanguage()));
    connect(m_destLanguageButton, SIGNAL(clicked()), this, SLOT(changeDestinationLanguage()));

    QTextCodec::setCodecForCStrings(QTextCodec::codecForName("UTF-8"));
    setAspectRatioMode(Plasma::IgnoreAspectRatio);
}

//========================================================================================

KTranslatoid ::~KTranslatoid()
{
    writeConfig();

}

//====================================================================================
void KTranslatoid::init()
{
    constructForm();
    readConfig();
    setupPopupIcon();
}

//=============
void KTranslatoid::constructForm()
{
    m_graphicsWidget = new QGraphicsWidget(this);

    Plasma::Frame *header = new Plasma::Frame(this);
    header->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Minimum);
    header->setText(i18n("Translatoid"));
    header->setAcceptHoverEvents(true);

    Plasma::ToolTipContent data;
    data.setMainText(i18n("Translatoid"));
    data.setSubText(i18n("Use the mouse wheel on icon to swap languages"));

    data.setImage(KIcon("translator").pixmap(IconSize(KIconLoader::Desktop)));

    Plasma::ToolTipManager::self()->setContent(this, data);

    m_buttonChange->setIcon(KIcon("system-switch-user.png"));
    m_buttonChange->nativeWidget()->setToolTip(i18n("Invert the translation"));

    m_buttonTranslate->setIcon(KIcon("tools-wizard.png"));
    m_buttonTranslate->setText(i18n("Translate"));
    m_buttonTranslate->nativeWidget()->setToolTip(i18n("Click here to translate"));

    QFont font = Plasma::Theme::defaultTheme()->font(Plasma::Theme::DefaultFont);

    m_sourceText->setFont(font);
    m_destText->setFont(font);

    m_destText->nativeWidget()->setReadOnly(true);

    Plasma::ToolButton * pastButton = new Plasma::ToolButton(this);
    Plasma::ToolButton * copyButton = new Plasma::ToolButton(this);
    Plasma::ToolButton * clearButton = new Plasma::ToolButton(this);
    Plasma::ToolButton * voiceButton = new Plasma::ToolButton(this);

    voiceButton->nativeWidget()->setIcon(KIcon("text-speak.png"));
    m_buttonChange->setMaximumSize(QSize(20,18));
    m_sourceLanguageButton->setMaximumSize(QSize(24,18));
    m_destLanguageButton->setMaximumSize(QSize(24,18));
    //    m_buttonTranslate->setMaximumSize(QSize(60, 20));

    voiceButton->setMaximumSize(QSize(20, 20));
    pastButton->setMaximumSize(QSize(20, 20));
    copyButton->setMaximumSize(QSize(20, 20));
    clearButton->setMaximumSize(QSize(20, 20));


    pastButton->nativeWidget()->setIcon(KIcon("edit-paste.png"));
    copyButton->nativeWidget()->setIcon(KIcon("edit-copy.png"));
    clearButton->nativeWidget()->setIcon(KIcon("edit-delete.png"));

    pastButton->nativeWidget()->setToolTip(i18n("Paste a selection to translate it"));
    copyButton->nativeWidget()->setToolTip(i18n("Copy the result"));
    clearButton->nativeWidget()->setToolTip(i18n("Clear the source text"));

    m_sourceLanguageButton->nativeWidget()->setToolTip(i18n("Change source language"));
    m_destLanguageButton->nativeWidget()->setToolTip(i18n("Change destination language"));

    connect(pastButton->nativeWidget(), SIGNAL(clicked()), this, SLOT(paste()));
    connect(copyButton->nativeWidget(), SIGNAL(clicked()), this, SLOT(copy()));
    connect(clearButton->nativeWidget(), SIGNAL(clicked()), m_sourceText->nativeWidget(), SLOT(clear()));
    connect(voiceButton->nativeWidget(), SIGNAL(clicked()), this, SLOT(tts()));

    QGraphicsLinearLayout*  layoutControl = new QGraphicsLinearLayout(Qt::Horizontal);

    layoutControl->addItem(m_sourceLanguageButton);
    layoutControl->addItem(m_buttonChange);
    layoutControl->addItem(m_destLanguageButton);
    layoutControl->addStretch();
    layoutControl->addItem(voiceButton);
    layoutControl->addItem(clearButton);
    layoutControl->addItem(pastButton);
    layoutControl->addItem(copyButton);
    layoutControl->addItem(m_buttonTranslate);

    layoutControl->setMaximumHeight(20);

    QGraphicsLinearLayout* mainLayout = new QGraphicsLinearLayout(Qt::Vertical);

    mainLayout->addItem(header);

    mainLayout->addItem(m_sourceLanguageLabel);
    mainLayout->addItem(m_sourceText);
    mainLayout->addItem(m_destLanguageLabel);
    mainLayout->addItem(m_destText);

    mainLayout->addItem(layoutControl);
    QColor col = Plasma::Theme::defaultTheme()->color(Plasma::Theme::defaultTheme()->TextColor);
    m_sourceText->setStyleSheet("color:"+col.name());
    m_destText->setStyleSheet("color:"+col.name());

    QGraphicsWidget * widget = new QGraphicsWidget(this);
    widget->setLayout(mainLayout);

    
    bigLayout = new QGraphicsLinearLayout(Qt::Horizontal);
    bigLayout->addItem(widget);
    m_graphicsWidget->setLayout(bigLayout);
    m_graphicsWidget->setPreferredSize(530, 320);

    m_sourceText->nativeWidget()->selectAll();
    m_sourceText->nativeWidget()->setHtml(i18n("Write here your text"));

    if (Solid::Networking::status() != Solid::Networking::Connected) {
        kDebug() << i18n("Networking is disabled. Cannot connect to google");
    }
}

//====================================================================================
QGraphicsWidget *KTranslatoid::graphicsWidget()
{

    //Explain : when the user change the plasma theme, need to update color Text by defaut theme. Otherwise, the text color can be same
    // than background color.
    QColor col = Plasma::Theme::defaultTheme()->color(Plasma::Theme::defaultTheme()->TextColor);
    m_sourceText->setStyleSheet("color:"+col.name());
    m_destText->setStyleSheet("color:"+col.name());
    return m_graphicsWidget;
}

//=============================================================================
void KTranslatoid::translate()
{
    kDebug() << "Starting translation.";

    KUrl url("http://www.google.com/translate_a/t");
    //kDebug() << "Url configured." << url;

    QByteArray postData;
    QString source = m_sourceLanguage.code.toUtf8().toPercentEncoding();
    QString dest = m_destLanguage.code.toUtf8().toPercentEncoding();
    QString texte= m_sourceText->nativeWidget()->toPlainText().toUtf8().toPercentEncoding();
    QString st = "text="+texte+"&client=t&sl=" + source + "&tl=" + dest;

    postData = st.toUtf8();

    
    if (m_job) {
        disconnect(m_job, 0, this, 0);
    }

    m_data.clear();
    delete m_job;

    m_job = KIO::http_post(url, postData, KIO::HideProgressInfo);
    m_job->addMetaData("Content-Type", "application/x-www-form-urlencoded");
    m_job->addMetaData("Host", "www.google.com");
    m_job->addMetaData("User-Agent", "Mozilla/5.0");
    m_job->addMetaData("Accept-Encoding", "deflate");
    m_job->addMetaData("Connection", "Close");

    connect(m_job, SIGNAL(data(KIO::Job*,QByteArray)), this, SLOT(httpDataReceived(KIO::Job*,QByteArray)));
    connect(m_job, SIGNAL(result(KJob*)), this, SLOT(done()));
    setBusy(true);
    kDebug() << "Request started";
    kDebug()<<postData;
}

//==============================================================================
void KTranslatoid::httpDataReceived(KIO::Job *job, QByteArray data)
{
    kDebug() << "got data for" << m_job << job << data;
    if (job != m_job) {
        return;
    }

    m_data.append(data);
}

//==============================================================================
void KTranslatoid::done()
{
    setBusy(false);
    m_job = 0;
    //kDebug() << "HTTP Request done." << m_data;


    QString text;
    text = text.fromUtf8(m_data);
    m_data.clear();
    kDebug() << "All data is read.";

    
    kDebug()<<text;

    text = text.replace(QString("\\\""),QString("\""));
    text = text.replace(QString("\\n"),QString("\n"));
    text = text.replace(QString("\n "),QString("\n"));
    text = text.replace(QString("\\x3c"),QString("<"));
    text = text.replace(QString("\\x3e"),QString(">"));

    if ( text.startsWith( QString("\"") ) ) {
        kDebug() << "This is a text.";
        text = text.remove( text.length()-1, 1).remove(0,1);
    } else if ( text.startsWith( QString("[") ) && text.endsWith( QString("]") ) ) {
        kDebug() << "This is a word.";
        QStringList tra;
        //tra = text.split(QRegExp(QString("\"(.*)\"")));
        text = text.replace(QString("]"),QString(""));
        tra = text.split(QString("["));
        text = QString("");
        bool languageIncluded = false;
        for (int i=0,j=0;i<tra.count();i++)
        {
            if (tra.at(i)!="")
            {
                if (j==0)
                {
                    QStringList translation = tra.at(i).split("\",\"");
                    if (translation.isEmpty()) {
                        continue;
                    }

                    text.append("<b>" + translation[0].replace("\"","") + "</b> ");

                    if (translation.count() > 1) {
                        languageIncluded = true;
                        QString langCode = translation[1].replace("\"","");
                        langCode = langCode.replace(",","");
                        Language la = m_languageModel->languageFromCode(langCode);
                        text.append("<i>(" + (la.name.isEmpty() ? langCode : la.name));
                        //kDebug() << "language code is" << langCode << la.name << text;
                    }
                } else {
                    QString translation = tra.at(i);
                    QStringList translations = translation.split(",");
                    for (int y = 0;y < translations.count(); ++y)
                    {
                        translation = translations.at(y);
                        translation = translation.replace("\"","");
                        if (y == 0)
                        {
                            if (languageIncluded) {
                                text.append(", ");
                            } else {
                                text.append("(<i>");
                            }

                            text.append(translation + "</i>)<br>");
                        } else {
                            text.append( QString( "\t" + translation + "\n" ) );
                        }
                    }
                    text.append( "\n" );
                }
                j++;
            }
        }
    }

    //kDebug() << "All operations done. Setting text to textarea.";
    m_destText->setText( text );
    //kDebug() << "Done.";
    //kDebug() << "Any else?";
}

//===================================================================================
void KTranslatoid::swapLanguages()
{
    updateLanguage(m_destLanguage, m_sourceLanguage);
    writeConfig();
}

//====================================================================================
void KTranslatoid::changeSourceLanguage()
{
    QMenu languageMenu;

    foreach (const Language &lang, m_languageModel->allLanguages()) {
        QAction *a = languageMenu.addAction(KIcon(lang.icon), lang.name);
        a->setData(lang.code);
    }

    connect(&languageMenu, SIGNAL(triggered(QAction*)),
            this, SLOT(sourceLanguageSelected(QAction*)));
    languageMenu.adjustSize();
    languageMenu.exec(containment()->corona()->popupPosition(m_sourceLanguageButton, languageMenu.size()));
}

//====================================================================================
void KTranslatoid::sourceLanguageSelected(QAction *action)
{
    if (!action) {
        return;
    }

    updateLanguage(m_languageModel->languageFromCode(action->data().toString()), m_destLanguage);
}

//====================================================================================
void KTranslatoid::changeDestinationLanguage()
{
    QMenu languageMenu;

    foreach (const Language &lang, m_languageModel->allLanguages()) {
        QAction *a = languageMenu.addAction(KIcon(lang.icon), lang.name);
        a->setData(lang.code);
    }

    connect(&languageMenu, SIGNAL(triggered(QAction*)),
            this, SLOT(destinationLanguageSelected(QAction*)));
    languageMenu.adjustSize();
    languageMenu.exec(containment()->corona()->popupPosition(m_sourceLanguageButton, languageMenu.size()));
}

//====================================================================================
void KTranslatoid::destinationLanguageSelected(QAction *action)
{
    if (!action) {
        return;
    }

    updateLanguage(m_sourceLanguage, m_languageModel->languageFromCode(action->data().toString()));
}
//=======================================================================================
void KTranslatoid::copy()
{
    m_destText->nativeWidget()->selectAll();
    m_destText->nativeWidget()->copy();
}
//=========================================================================================
void KTranslatoid::paste()
{
    m_sourceText->nativeWidget()->clear();
    m_sourceText->nativeWidget()->paste();
}
//==========================================================================================
void KTranslatoid::tts()
{

    QString qcommand = m_command;
    qcommand.replace("$RESULT", m_destText->nativeWidget()->toPlainText());
    qcommand.replace("$SOURCE", m_sourceText->nativeWidget()->toPlainText());


    QByteArray ba = qcommand.toUtf8();
    const char *command = ba.data();
    if (!system(command))
    {
        kDebug() <<"Unable to call TTS program";
        //        KMessageBox::error(0,i18n("you need to install festival"),i18n("cannot find tts support"));

    }
}

//==============================================================================================
void KTranslatoid::createConfigurationInterface(KConfigDialog *parent)
{
    m_voiceForm = new VoiceForm;
    QWidget *languageForm = new QWidget;
    m_ui.setupUi(languageForm);
    m_ui.checkBoxPaste->setChecked(m_autoPaste);
    m_ui.checkBoxTranslate->setChecked(m_autoTranslate);

    parent->addPage(languageForm, i18n("Translation"), "preferences-desktop-locale");
    parent->addPage(m_voiceForm, i18n("Voice"), "preferences-desktop-text-to-speech");

    connect(parent, SIGNAL(applyClicked()), this, SLOT(configAccepted()));
    connect(parent, SIGNAL(okClicked()), this, SLOT(configAccepted()));
}

//================================================================================================
void KTranslatoid::configAccepted()
{
    m_command = m_voiceForm->commandLine();
    m_autoPaste = m_ui.checkBoxPaste->isChecked();
    m_autoTranslate = m_ui.checkBoxTranslate->isChecked();
    writeConfig();
}

//============================================================================================================
void KTranslatoid::popupEvent(bool show)
{

    if (show) {
        m_sourceText->grabKeyboard();
        if (m_autoPaste) {
            QClipboard *clipboard = QApplication::clipboard();


            m_sourceText->nativeWidget()->clear();
            m_sourceText->nativeWidget()->setText(clipboard->text(QClipboard::Selection));


            //             m_sourceText->nativeWidget()->paste();
        }

        if (m_autoTranslate)
            translate();

    }

    else {
        m_sourceText->ungrabKeyboard();

    }

}
//============================================================================================================
void KTranslatoid::setupPopupIcon()
{
    QPixmap pix(64, 64);
    pix.fill(Qt::transparent);
    QPainter paint(&pix);

    paint.drawPixmap(0, 0, KIcon("translator").pixmap(64, 64));

    QPixmap flag1 = KIcon(m_sourceLanguage.icon).pixmap(24, 33);
    QPixmap flag2 = KIcon(m_destLanguage.icon).pixmap(24, 33);
    //
    //    QTransform transform;
    //    transform.shear(0.15, -0.15);
    //    flag1 = flag1.transformed(transform, Qt::SmoothTransformation);
    //    flag2 = flag2.transformed(transform, Qt::SmoothTransformation);


    paint.drawPixmap(0, 40, flag1);
    paint.drawPixmap(40, 0, flag2);
    //    paint.setTransform(transform, Qt::SmoothTransformation);
    paint.setPen(Qt::white);

    //     paint.drawText(QRect(10,35,pix.width(),pix.height()), m_sourceLanguage.code+"  "+m_destLanguage.code);
    m_popupIcon = KIcon(pix);

    setPopupIcon(m_popupIcon);

}
//===============================================
void KTranslatoid::updateLanguage(Language a, Language b)
{
    m_sourceLanguage = a;
    m_destLanguage = b;
    m_sourceLanguageButton->nativeWidget()->setIcon(KIcon(m_sourceLanguage.icon));
    m_sourceLanguageLabel->setText(m_sourceLanguage.name);

    m_destLanguageButton->nativeWidget()->setIcon(KIcon(m_destLanguage.icon));
    m_destLanguageLabel->setText(m_destLanguage.name);

    setupPopupIcon();
}

//================================================
void KTranslatoid::readConfig()
{
    KConfigGroup cg = config();

    QString a = cg.readEntry("lsource", QString("fr"));
    QString b = cg.readEntry("lresult", QString("en"));
    m_autoPaste = cg.readEntry("autoPaste", true);
    m_autoTranslate = cg.readEntry("autoTranslate", true);

    Language la = m_languageModel->languageFromCode(a);
    Language lb = m_languageModel->languageFromCode(b);

    updateLanguage(la, lb);

    QString defaultCommand;
    if (!KStandardDirs::findExe("festival").isEmpty()) {
        defaultCommand = "echo '$RESULT'|festival --tts";
    } else if (!KStandardDirs::findExe("espeak").isEmpty()) {
        defaultCommand = "espeak '$RESULT'";
    }
    m_command = cg.readEntry("tts", defaultCommand);
}

void KTranslatoid::writeConfig()
{
    KConfigGroup cg = config();

    cg.writeEntry("lsource", m_sourceLanguage.code);
    cg.writeEntry("lresult", m_destLanguage.code);

    cg.writeEntry("tts", m_command);
    cg.writeEntry("autoPaste", m_autoPaste);
    cg.writeEntry("autoTranslate", m_autoTranslate);

    emit configNeedsSaving();
}

bool KTranslatoid::eventFilter(QObject * obj, QEvent *event)
{
    //Filter TextEdit Event
    //Enter  = translate
    //Shift+Enter = add new line
    if (obj == m_sourceText->nativeWidget()) {
        if (event->type() == QEvent::KeyPress) {
            QKeyEvent *keyEvent = static_cast<QKeyEvent*>(event);

            if (keyEvent->key() == Qt::Key_Return && keyEvent->modifiers() == Qt::ShiftModifier) {
                m_sourceText->nativeWidget()->append("");
                return true;
            }

            if (keyEvent->key() == Qt::Key_Return || keyEvent->key() == Qt::Key_Enter) {
                translate();
                return true;
            }
        }
    }

    return Plasma::PopupApplet::eventFilter(obj, event);
}

//================================================================
void KTranslatoid::wheelEvent(QGraphicsSceneWheelEvent *event)
{
    Q_UNUSED(event);
    swapLanguages();
}

#include "ktranslatoid.moc"

